@@ -28,6 +28,7 @@ import multiprocessing
 import operator
 import os
 import os.path as osp
+import time
 import tokenize
 import types
 import platform
@@ -552,11 +553,13 @@ def profile_runs(funcname, modulefile, nruns=5):
     return stats
 
 
-def save_plist(cfg_file, my_map):
-    with open(cfg_file, 'rb') as fp:
+def save_plist(path, my_map):
+    with open(path, 'rb') as fp:
         plist = plistlib.load(fp, fmt=plistlib.FMT_XML)
     plist.update(my_map)
-    with open(cfg_file, 'wb') as fp:
+    par_dir = osp.split(path)[0]
+    os.makedirs(par_dir, exist_ok=True)
+    with open(path, 'wb') as fp:
         plistlib.dump(plist, fp)
 
 
@@ -765,6 +768,34 @@ def rerun_lock(name, folder=None, logger=glogger):
     return decorator
 
 
+def await_condition(condition, timeout_ms, step_ms=10):
+    """
+    - condition must implement methods:
+      - .met()
+      - .update()
+    - under concurrency, caller must ensure thread/process safety, e.g., implement locking in condition
+    """
+    waited_ms = 0
+    while waited_ms < timeout_ms and condition.met():
+        condition.update()
+        time.sleep(step_ms/1000)
+        waited_ms += step_ms
+    return waited_ms < timeout_ms
+
+
+def await_lockfile(lockpath, timeout_ms=float('inf'), step_ms=10):
+    class PathExistsCondition:
+        def __init__(self, path):
+            self.path = path
+
+        def met(self):
+            return osp.exists(self.path)
+
+        def update(self):
+            pass
+    return await_condition(PathExistsCondition(lockpath), timeout_ms, step_ms)
+
+
 def append_to_os_paths(bindir, usesyspath=True, inmemonly=False):
     """
     On macOS, PATH update will only take effect after calling `source ~/.bash_profile` directly in shell. It won't work 
@@ -1409,7 +1440,7 @@ def compare_dirs(dir1, dir2, ignoreddirpatterns=(), ignoredfilepatterns=(), show
     assert dir1_contents['files'] == dir2_contents['files'], 'files different:\n{}\n\nvs.\n\n{}'.format(pp.pformat(dir1_contents['files'], indent=2), pp.pformat(dir2_contents['files'], indent=2))
 
 
-def pack_obj(obj, topic=None, envelope=('<KK-ENV>', '</KK-ENV>'), classes=()):
+def pack_obj(obj, topic=None, envelope=('<KK-ENV>', '</KK-ENV>'), classes=(), ensure_ascii=False):
     """
     for cross-language rpc only, so no need for an unpack()
     pickle is already good enough for local transmission.
@@ -1426,9 +1457,9 @@ def pack_obj(obj, topic=None, envelope=('<KK-ENV>', '</KK-ENV>'), classes=()):
                 else:
                     return json.JSONEncoder.encode(self, o)
         classes = tuple([SimpleNamespace]+list(classes))
-        msg_str = json.dumps(msg, cls=CustomJsonEncoder)
+        msg_str = json.dumps(msg, cls=CustomJsonEncoder, ensure_ascii=ensure_ascii)
     else:
-        msg_str = json.dumps(msg, default=lambda o: o.__dict__)
+        msg_str = json.dumps(msg, default=lambda o: o.__dict__, ensure_ascii=ensure_ascii)
 
     return f'{envelope[0]}{msg_str}{envelope[1]}'
 
@@ -1496,6 +1527,58 @@ def get_child_dirs(root, subs=()):
     return (osp.join(root, sub) for sub in subs)
 
 
+def get_drivewise_commondirs(paths: list[str]):
+    """
+    - windows paths are drive-bound
+    - paths must be either all absolute or all relative
+    - on windows, we distinguish b/w UNC and relative paths
+    - on posix, drive letter is always '' (empty str) for absolute and relative paths
+      - so we differentiate them for clarity: '/' and ''
+    - posix: common dir of below: a single drive map: {'/': 'path/to'}
+      - /path/to/dir1/file1
+      - /path/to/dir2/
+      - /path/to/dir3/dir4/file2
+    - posix: common dir of below: a single drive map: {'': 'path/to'}
+      - path/to/dir1/file1
+      - path/to/dir2/
+      - path/to/dir3/dir4/file2
+    - windows: common dirs of below: a multi-drive map: {'c:': 'path\\to', 'd:': 'path\\to', '\\\\network\\share': 'path\\to\\dir7', '': 'path\\to\\dir9'}
+      - C:\\path\\to\\dir1\\file1
+      - C:\\path\\to\\dir2\\
+      - D:\\path\\to\\dir3\\dir4\\file2
+      - D:\\path\\to\\dir5\\dir6\\file3
+      - \\\\network\\share\\path\\to\\dir7\\file4
+      - \\\\network\\share\\path\\to\\dir7\\dir8\\file5
+      - path\\to\\dir9\\file6
+      - path\\to\\dir9\\file7
+    - on windows, we normalize all paths to lowercase
+    """
+    if is_posix := platform.system() != 'Windows':
+        single_cm_dir = osp.commonpath(paths)
+        if len(paths) == 1:
+            single_cm_dir = osp.dirname(single_cm_dir)
+        root = '/'
+        drive = root if single_cm_dir.startswith(root) else ''
+        return {drive: single_cm_dir.strip(root)}
+    # windows
+    if len(paths) == 1:
+        drive, relpath = osp.splitdrive(paths[0])
+        drive = drive.lower()
+        single_cm_dir = osp.dirname(relpath).strip('\\')
+        return {drive: single_cm_dir}
+    paths_sorted_by_drive = sorted(paths)
+    # collect paths into map by drive
+    drive_path_map = {}
+    for p, winpath in enumerate(paths_sorted_by_drive):
+        assert winpath, f'Invalid path at line {p}; must not be empty'
+        drive, relpath = osp.splitdrive(winpath)
+        drive = drive.lower()
+        if must_lazy_init_for_drive := drive not in drive_path_map:
+            drive_path_map[drive] = []
+        drive_path_map[drive].append(relpath.strip('\\'))
+    return {drive: osp.dirname(winpaths[0]).strip('\\') if (is_single_file := len(winpaths) == 1 and osp.splitdrive(winpaths[0])[1]) else osp.commonpath(winpaths).strip('\\') for drive, winpaths in drive_path_map.items()}
+
+
 def read_lines(file, striplineend=False, posix=True):
     with open(file) as fp:
         lines = fp.readlines()
@@ -1635,6 +1718,8 @@ def save_lines(path, lines, toappend=False, addlineend=False, style='posix'):
     if addlineend:
         line_end = '\n' if style == 'posix' else '\r\n'
         lines_to_write = [line+line_end for line in lines]
+    par_dir = osp.split(path)[0]
+    os.makedirs(par_dir, exist_ok=True)
     with open(path, mode) as fp:
         fp.writelines(lines_to_write)
     return lines_to_write
@@ -1646,9 +1731,10 @@ def load_text(path):
     return text
 
 
-def save_text(path, text, toappend=False):
-    assert isinstance(text, str)
+def save_text(path, text: str, toappend=False):
     mode = 'a' if toappend else 'w'
+    par_dir = osp.split(path)[0]
+    os.makedirs(par_dir, exist_ok=True)
     with open(path, mode) as fp:
         fp.write(text)
 
